# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Web server implementation..

This is the concrete web server implementation. It provides the
high level application logic that combines the WebDAV server,
the carddav support, the caldav support and the DAV store.
"""

import asyncio
import functools
import hashlib
import logging
from logging import getLogger
import os
import posixpath
import shutil
import socket
import urllib.parse
from collections.abc import Iterable, Iterator
from email.utils import parseaddr
from dulwich.web import make_wsgi_chain
from dulwich.server import DictBackend
from itertools import takewhile

import jinja2

from xandikos import __version__ as xandikos_version
from xandikos import (
    access,
    apache,
    caldav,
    carddav,
    infit,
    quota,
    scheduling,
    sync,
    timezones,
    webdav,
    xmpp,
)
from xandikos.fs import FilesystemBackend, open_store_from_path
from xandikos.store import (
    STORE_TYPE_ADDRESSBOOK,
    STORE_TYPE_CALENDAR,
    STORE_TYPE_OTHER,
    STORE_TYPE_PRINCIPAL,
    STORE_TYPE_SCHEDULE_INBOX,
    STORE_TYPE_SCHEDULE_OUTBOX,
    STORE_TYPE_SUBSCRIPTION,
    DuplicateUidError,
    File,
    InvalidCTag,
    InvalidFileContents,
    LockedError,
    NoSuchItem,
    NotStoreError,
    OutOfSpaceError,
    Store,
)

from .icalendar import CalendarFilter
from .store.git import GitStore, TreeGitStore

logger = getLogger("xandikos")

try:
    import systemd.daemon
except ImportError:
    systemd_imported = False

    def get_systemd_listen_sockets() -> list[socket.socket]:
        raise NotImplementedError
else:
    systemd_imported = True

    def get_systemd_listen_sockets() -> list[socket.socket]:
        socks = []
        for fd in systemd.daemon.listen_fds():
            for family in (
                socket.AF_UNIX,  # type: ignore
                socket.AF_INET,
                socket.AF_INET6,
            ):
                if systemd.daemon.is_socket(
                    fd, family=family, type=socket.SOCK_STREAM, listening=True
                ):
                    sock = socket.fromfd(fd, family, socket.SOCK_STREAM)
                    socks.append(sock)
                    break
            else:
                raise RuntimeError(
                    "socket family must be AF_INET, AF_INET6, or AF_UNIX; "
                    "socket type must be SOCK_STREAM; and it must be listening"
                )
        return socks


WELLKNOWN_DAV_PATHS = {
    caldav.WELLKNOWN_CALDAV_PATH,
    carddav.WELLKNOWN_CARDDAV_PATH,
}

# TODO(jelmer): Make these configurable/dynamic
CALENDAR_HOME_SET = ["calendars"]
ADDRESSBOOK_HOME_SET = ["contacts"]
GIT_PATH = ".git"

# Mapping from content types to their validation error tags
CONTENT_TYPE_ERROR_TAGS = {
    "text/calendar": ("{%s}valid-calendar-data" % caldav.NAMESPACE, "calendar"),
    "text/vcard": ("{%s}valid-address-data" % carddav.NAMESPACE, "vCard"),
}


def get_validation_error(exc: InvalidFileContents):
    """Get appropriate validation error tag for a content type.

    Args:
        exc: InvalidFileContents exception with content_type and error details

    Returns:
        Tuple of (error_tag, error_message) for the content type
    """
    error_tag, file_type = CONTENT_TYPE_ERROR_TAGS.get(
        exc.content_type,
        ("{%s}valid-calendar-data" % caldav.NAMESPACE, "file"),
    )
    return error_tag, f"Not a valid {file_type} file: {exc.error}"


TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), "templates")
jinja_env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(TEMPLATES_DIR), enable_async=True
)


async def render_jinja_page(
    name: str, accepted_content_languages: list[str], **kwargs
) -> tuple[Iterable[bytes], int, str | None, str, list[str]]:
    """Render a HTML page from jinja template.

    Args:
      name: Name of the page
      accepted_content_languages: List of accepted content languages
    Returns: Tuple of (body, content_length, etag, content_type, languages)
    """
    # TODO(jelmer): Support rendering other languages
    encoding = "utf-8"
    template = jinja_env.get_template(name)
    body = await template.render_async(
        version=xandikos_version, urljoin=urllib.parse.urljoin, **kwargs
    )
    body_encoded = body.encode(encoding)
    return (
        [body_encoded],
        len(body_encoded),
        None,
        f"text/html; encoding={encoding}",
        ["en-UK"],
    )


def create_strong_etag(etag: str) -> str:
    """Create strong etags.

    Args:
      etag: basic etag
    Returns: A strong etag
    """
    return '"' + etag + '"'


def extract_strong_etag(etag: str | None) -> str | None:
    """Extract a strong etag from a string."""
    if etag is None:
        return etag
    return etag.strip('"')


class ObjectResource(webdav.Resource):
    """Object resource."""

    def __init__(
        self,
        store: Store,
        name: str,
        content_type: str,
        etag: str,
        file: File | None = None,
    ) -> None:
        self.store = store
        self.name = name
        self.etag = etag
        self.content_type = content_type
        self._file = file

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.store!r}, {self.name!r}, {self.etag!r}, {self.get_content_type()!r})"

    async def get_file(self) -> File:
        if self._file is None:
            self._file = await asyncio.to_thread(
                self.store.get_file, self.name, self.content_type, self.etag
            )
            assert self._file is not None
        return self._file

    async def get_body(self) -> Iterable[bytes]:
        file = await self.get_file()
        return file.content

    async def set_body(self, data, replace_etag=None, remote_user=None, requester=None):
        try:
            (name, etag) = await asyncio.to_thread(
                self.store.import_one,
                self.name,
                self.content_type,
                data,
                replace_etag=extract_strong_etag(replace_etag),
                remote_user=remote_user,
                requester=requester,
            )
        except InvalidFileContents as exc:
            error_tag, error_message = get_validation_error(exc)
            raise webdav.PreconditionFailure(error_tag, error_message) from exc
        except DuplicateUidError as exc:
            raise webdav.PreconditionFailure(
                "{%s}no-uid-conflict" % caldav.NAMESPACE, "UID already in use."
            ) from exc
        except LockedError as exc:
            raise webdav.ResourceLocked() from exc
        return create_strong_etag(etag)

    def get_content_language(self) -> str:
        raise KeyError

    def get_content_type(self) -> str:
        return self.content_type

    async def get_content_length(self) -> int:
        return sum(map(len, await self.get_body()))

    async def get_etag(self) -> str:
        return create_strong_etag(self.etag)

    def get_supported_locks(self):
        return []

    def get_active_locks(self):
        return []

    def get_owner(self):
        return None

    def get_comment(self):
        raise KeyError

    def set_comment(self, comment):
        raise NotImplementedError(self.set_comment)

    def get_creationdate(self):
        # TODO(jelmer): Find creation date using store function
        raise KeyError

    def get_last_modified(self):
        # TODO(jelmer): Find last modified time using store function
        raise KeyError

    def get_is_executable(self):
        # TODO(jelmer): Retrieve POSIX mode and check for executability.
        return False

    def get_quota_used_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError

    def get_quota_available_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError

    def get_schedule_tag(self):
        # TODO(jelmer): Ask the store?
        raise KeyError


class StoreBasedCollection:
    def __init__(self, backend, relpath, store) -> None:
        self.backend = backend
        self.relpath = relpath
        self.store = store

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.store!r})"

    def set_resource_types(self, resource_types):
        # TODO(jelmer): Allow more than just this set; allow combining
        # addressbook/calendar.
        resource_types = set(resource_types)
        if resource_types == {
            caldav.CALENDAR_RESOURCE_TYPE,
            webdav.COLLECTION_RESOURCE_TYPE,
        }:
            self.store.set_type(STORE_TYPE_CALENDAR)
        elif resource_types == {
            carddav.ADDRESSBOOK_RESOURCE_TYPE,
            webdav.COLLECTION_RESOURCE_TYPE,
        }:
            self.store.set_type(STORE_TYPE_ADDRESSBOOK)
        elif resource_types == {webdav.PRINCIPAL_RESOURCE_TYPE}:
            self.store.set_type(STORE_TYPE_PRINCIPAL)
        elif resource_types == {
            caldav.SCHEDULE_INBOX_RESOURCE_TYPE,
            webdav.COLLECTION_RESOURCE_TYPE,
        }:
            self.store.set_type(STORE_TYPE_SCHEDULE_INBOX)
        elif resource_types == {
            caldav.SCHEDULE_OUTBOX_RESOURCE_TYPE,
            webdav.COLLECTION_RESOURCE_TYPE,
        }:
            self.store.set_type(STORE_TYPE_SCHEDULE_OUTBOX)
        elif resource_types == {webdav.COLLECTION_RESOURCE_TYPE}:
            self.store.set_type(STORE_TYPE_OTHER)
        elif resource_types == {
            webdav.COLLECTION_RESOURCE_TYPE,
            caldav.SUBSCRIPTION_RESOURCE_TYPE,
        }:
            self.store.set_type(STORE_TYPE_SUBSCRIPTION)
        else:
            raise NotImplementedError(self.set_resource_types)

    def _get_resource(
        self,
        name: str,
        content_type: str,
        etag: str,
        file: File | None = None,
    ) -> webdav.Resource:
        return ObjectResource(self.store, name, content_type, etag, file=file)

    def _get_subcollection(self, name: str) -> webdav.Collection:
        return self.backend.get_resource(posixpath.join(self.relpath, name))

    def get_displayname(self) -> str:
        displayname = self.store.get_displayname()
        if displayname is None:
            return os.path.basename(self.store.repo.path)
        return displayname

    def set_displayname(self, displayname: str) -> None:
        self.store.set_displayname(displayname)

    def get_sync_token(self) -> str:
        return self.store.get_ctag()

    def get_ctag(self) -> str:
        return self.store.get_ctag()

    async def get_etag(self) -> str:
        return create_strong_etag(self.store.get_ctag())

    def members(self) -> Iterator[tuple[str, webdav.Resource]]:
        for name, content_type, etag in self.store.iter_with_etag():
            resource = self._get_resource(name, content_type, etag)
            yield (name, resource)
        for name, resource in self.subcollections():
            yield (name, resource)

    def subcollections(self):
        for name in self.store.subdirectories():
            yield (name, self._get_subcollection(name))

    def get_member(self, name):
        assert name != ""
        try:
            (content_type, etag) = self.store.get_file_meta(name)
        except KeyError:
            if name in self.store.subdirectories():
                return self._get_subcollection(name)
            raise KeyError(name)
        return self._get_resource(name, content_type, etag)

    def delete_member(self, name, etag=None, remote_user=None, requester=None):
        assert name != ""
        try:
            self.store.delete_one(
                name,
                etag=extract_strong_etag(etag),
                remote_user=remote_user,
                requester=requester,
            )
        except NoSuchItem:
            try:
                _subcoll = self._get_subcollection(name)
            except KeyError:
                # Item doesn't exist at all, raise KeyError to return 404
                raise KeyError(name)
            else:
                # TODO: Properly allow removing subcollections
                # _subcoll.destroy()
                shutil.rmtree(os.path.join(self.store.path, name))

    async def create_member(
        self,
        name: str,
        contents: Iterable[bytes],
        content_type: str,
        remote_user: str | None = None,
        requester: str | None = None,
    ) -> tuple[str, str]:
        # Check if member already exists and raise FileExistsError if it does
        try:
            existing_member = self.get_member(name)
            if existing_member is not None:
                raise FileExistsError(f"Member '{name}' already exists")
        except KeyError:
            # Member doesn't exist, which is what we want for create_member
            pass

        try:
            (name, etag) = self.store.import_one(
                name,
                content_type,
                contents,
                remote_user=remote_user,
                requester=requester,
            )
        except InvalidFileContents as exc:
            error_tag, error_message = get_validation_error(exc)
            raise webdav.PreconditionFailure(error_tag, error_message) from exc
        except DuplicateUidError as exc:
            raise webdav.PreconditionFailure(
                "{%s}no-uid-conflict" % caldav.NAMESPACE, "UID already in use."
            ) from exc
        except OutOfSpaceError as exc:
            raise webdav.InsufficientStorage() from exc
        except LockedError as exc:
            raise webdav.ResourceLocked() from exc
        return (name, create_strong_etag(etag))

    def iter_differences_since(
        self, old_token: str, new_token: str
    ) -> Iterator[tuple[str, webdav.Resource | None, webdav.Resource | None]]:
        old_resource: webdav.Resource | None
        new_resource: webdav.Resource | None
        try:
            for (
                name,
                content_type,
                old_etag,
                new_etag,
            ) in self.store.iter_changes(old_token, new_token):
                if old_etag is not None:
                    old_resource = self._get_resource(name, content_type, old_etag)
                else:
                    old_resource = None
                if new_etag is not None:
                    new_resource = self._get_resource(name, content_type, new_etag)
                else:
                    new_resource = None
                yield (name, old_resource, new_resource)
        except InvalidCTag as exc:
            raise sync.InvalidToken(exc.ctag) from exc

    def get_owner(self):
        return None

    def get_supported_locks(self):
        return []

    def get_active_locks(self):
        return []

    def get_headervalue(self):
        raise KeyError

    def get_comment(self):
        return self.store.get_comment()

    def set_comment(self, comment):
        self.store.set_comment(comment)

    def get_creationdate(self):
        # TODO(jelmer): Find creation date using store function
        raise KeyError

    def get_last_modified(self):
        # TODO(jelmer): Find last modified time using store function
        raise KeyError

    def get_content_type(self):
        return "httpd/unix-directory"

    def get_content_language(self):
        raise KeyError

    async def get_content_length(self):
        raise KeyError

    def destroy(self) -> None:
        # RFC2518, section 8.6.2 says this should recursively delete.
        self.store.destroy()
        self.backend._open_store.cache_clear()

    async def get_body(self):
        raise NotImplementedError(self.get_body)

    async def render(
        self, self_url, accepted_content_types, accepted_content_languages
    ):
        content_types = webdav.pick_content_types(accepted_content_types, ["text/html"])
        assert content_types == ["text/html"]
        return await render_jinja_page(
            "collection.html",
            accepted_content_languages,
            collection=self,
            self_url=self_url,
        )

    def get_is_executable(self) -> bool:
        return False

    def get_quota_used_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError

    def get_quota_available_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError

    def get_refreshrate(self):
        return self.store.config.get_refreshrate()

    def set_refreshrate(self, value):
        self.store.config.set_refreshrate(value)


class Collection(StoreBasedCollection, webdav.Collection):
    """A generic WebDAV collection."""


class ScheduleInbox(StoreBasedCollection, scheduling.ScheduleInbox):
    """A schedling inbox collection."""


class ScheduleOutbox(StoreBasedCollection, scheduling.ScheduleOutbox):
    """A schedling outbox collection."""


class SubscriptionCollection(StoreBasedCollection, caldav.Subscription):
    def get_source_url(self):
        source_url = self.store.get_source_url()
        if source_url is None:
            raise KeyError
        return source_url

    def set_source_url(self, url):
        self.store.set_source_url(url)

    def get_calendar_description(self):
        return self.store.get_description()

    def get_calendar_color(self):
        color = self.store.get_color()
        if not color:
            raise KeyError
        if color and color[0] != "#":
            color = "#" + color
        return color

    def set_calendar_color(self, color):
        self.store.set_color(color)

    def get_supported_calendar_components(self):
        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VAVAILABILITY"]


class CalendarCollection(StoreBasedCollection, caldav.Calendar):
    def get_calendar_description(self):
        return self.store.get_description()

    def set_calendar_description(self, description):
        self.store.set_description(description)

    def get_calendar_color(self):
        color = self.store.get_color()
        if not color:
            raise KeyError
        if color and color[0] != "#":
            color = "#" + color
        return color

    def set_calendar_color(self, color):
        self.store.set_color(color)

    def get_calendar_order(self):
        order = self.store.config.get_order()
        if not order:
            raise KeyError
        return order

    def set_calendar_order(self, order):
        self.store.config.set_order(order)

    def get_calendar_timezone(self):
        return self.store.config.get_timezone()

    def set_calendar_timezone(self, content):
        self.store.config.set_timezone(content)

    def _ensure_metadata_directory(self):
        """Ensure .xandikos/ metadata directory exists, migrating from old .xandikos config file if needed."""
        # Check if we already have the new directory structure by checking for config file
        try:
            self.store.get_file(".xandikos/config", "text/plain")
            return  # Already migrated
        except KeyError:
            pass  # Need to migrate or create

        # Check if we have the old .xandikos file that needs migration
        old_config_content = None
        try:
            old_config_file = self.store.get_file(".xandikos", "text/plain")
        except KeyError:
            pass  # No old config file to migrate
        else:
            old_config_content = b"".join(old_config_file.content)
            # Remove the old file
            self.store.delete_one(".xandikos")

        # Create .xandikos/ metadata directory by creating config file within it
        if old_config_content:
            # Migrate old config file content
            content = [old_config_content]
            message = "Migrate .xandikos config to metadata directory structure"
        else:
            # Create empty config file to establish the metadata directory
            content = [b""]
            message = "Create .xandikos metadata directory structure"

        self.store.import_one(
            ".xandikos/config", "text/plain", content, message=message
        )

    def get_calendar_availability(self):
        """Get calendar availability from .xandikos/availability.ics file."""
        try:
            availability_file = self.store.get_file(
                ".xandikos/availability.ics", "text/calendar"
            )
        except NoSuchItem:
            raise KeyError

        return b"".join(availability_file.content).decode("utf-8")

    def set_calendar_availability(self, content):
        """Set calendar availability by storing in .xandikos/availability.ics file."""
        # Ensure .xandikos/ metadata directory exists (migrates config if needed)
        self._ensure_metadata_directory()

        if content is None:
            # Remove availability
            try:
                self.store.delete_one(".xandikos/availability.ics")
            except NoSuchItem:
                pass  # Already removed
        else:
            # Validate that it's valid iCalendar data and normalize it
            try:
                from icalendar.cal import Calendar as ICalendar

                cal = ICalendar.from_ical(content)
            except (ValueError, UnicodeDecodeError, TypeError, KeyError) as e:
                raise InvalidFileContents("text/calendar", content, e)

            # Store the normalized form
            normalized_content = cal.to_ical().decode("utf-8")
            self.store.import_one(
                ".xandikos/availability.ics",
                "text/calendar",
                [normalized_content.encode("utf-8")],
                message="Update calendar availability",
            )

    def get_supported_calendar_components(self):
        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY", "VAVAILABILITY"]

    def get_supported_calendar_data_types(self):
        return [("text/calendar", "1.0"), ("text/calendar", "2.0")]

    def get_max_date_time(self):
        return "99991231T235959Z"

    def get_min_date_time(self):
        return "00010101T000000Z"

    def get_max_instances(self):
        raise KeyError

    def get_max_attendees_per_instance(self):
        raise KeyError

    def get_max_resource_size(self):
        # No resource limit
        raise KeyError

    def get_max_attachments_per_resource(self):
        # No resource limit
        raise KeyError

    def get_max_attachment_size(self):
        # No resource limit
        raise KeyError

    def get_schedule_calendar_transparency(self):
        # TODO(jelmer): Allow configuration in config
        return caldav.TRANSPARENCY_OPAQUE

    def get_managed_attachments_server_url(self):
        # TODO(jelmer)
        raise KeyError

    def calendar_query(self, create_filter_fn):
        filter = create_filter_fn(CalendarFilter)
        for name, file, etag in self.store.iter_with_filter(filter=filter):
            resource = self._get_resource(name, file.content_type, etag, file=file)
            yield (name, resource)

    def get_xmpp_heartbeat(self):
        # TODO
        raise KeyError

    def get_xmpp_server(self):
        # TODO
        raise KeyError

    def get_xmpp_uri(self):
        # TODO
        raise KeyError


class AddressbookCollection(StoreBasedCollection, carddav.Addressbook):
    def get_addressbook_description(self):
        return self.store.get_description()

    def set_addressbook_description(self, description):
        self.store.set_description(description)

    def get_supported_address_data_types(self):
        return [("text/vcard", "3.0")]

    def get_max_resource_size(self):
        # No resource limit
        raise KeyError

    def get_max_image_size(self):
        # No resource limit
        raise KeyError

    def set_addressbook_color(self, color):
        self.store.set_color(color)

    def addressbook_query(self, create_filter_fn):
        from .vcard import CardDAVFilter

        filter = create_filter_fn(CardDAVFilter)
        for name, file, etag in self.store.iter_with_filter(filter=filter):
            resource = self._get_resource(name, file.content_type, etag, file=file)
            yield (name, resource)

    def get_addressbook_color(self):
        color = self.store.get_color()
        if not color:
            raise KeyError
        if color and color[0] != "#":
            color = "#" + color
        return color


class CollectionSetResource(webdav.Collection):
    """Resource for calendar sets."""

    def __init__(self, backend, relpath) -> None:
        self.backend = backend
        self.relpath = relpath

    @classmethod
    def create(cls, backend, relpath):
        path = backend._map_to_file_path(relpath)
        if not os.path.isdir(path):
            os.makedirs(path)
            logger.info("Creating %s", path)
        return cls(backend, relpath)

    def get_displayname(self):
        return posixpath.basename(self.relpath)

    def get_sync_token(self):
        raise KeyError

    async def get_etag(self):
        raise KeyError

    def get_ctag(self):
        raise KeyError

    def get_supported_locks(self):
        return []

    def get_active_locks(self):
        return []

    def get_owner(self):
        return None

    def members(self):
        p = self.backend._map_to_file_path(self.relpath)
        for name in os.listdir(p):
            if name.startswith("."):
                continue
            resource = self.get_member(name)
            yield (name, resource)

    def get_member(self, name):
        assert name != ""
        relpath = posixpath.join(self.relpath, name)
        p = self.backend._map_to_file_path(relpath)
        if not os.path.isdir(p):
            raise KeyError(name)
        return self.backend.get_resource(relpath)

    def get_headervalue(self):
        raise KeyError

    def get_comment(self):
        raise KeyError

    def set_comment(self, comment):
        raise NotImplementedError(self.set_comment)

    def get_content_type(self):
        return "httpd/unix-directory"

    def get_content_language(self):
        raise KeyError

    async def get_content_length(self):
        raise KeyError

    def get_last_modified(self):
        # TODO(jelmer): Find last modified time using store function
        raise KeyError

    def delete_member(self, name, etag=None, remote_user=None, requester=None):
        # This doesn't have any non-collection members.
        self.get_member(name).destroy()

    def destroy(self):
        p = self.backend._map_to_file_path(self.relpath)
        # RFC2518, section 8.6.2 says this should recursively delete.
        shutil.rmtree(p)
        self.backend._open_store.cache_clear()

    async def render(
        self, self_url, accepted_content_types, accepted_content_languages
    ):
        content_types = webdav.pick_content_types(accepted_content_types, ["text/html"])
        assert content_types == ["text/html"]
        return await render_jinja_page(
            "root.html", accepted_content_languages, self_url=self_url
        )

    def get_is_executable(self):
        return False

    def get_quota_used_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError

    def get_quota_available_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError

    def get_creationdate(self):
        # TODO(jelmer): Find creation date using store function
        raise KeyError


class RootPage(webdav.Resource):
    """A non-DAV resource."""

    resource_types: list[str] = []

    def __init__(self, backend, show_principals: bool = True) -> None:
        self.backend = backend
        self.show_principals = show_principals

    def render(self, self_url, accepted_content_types, accepted_content_languages):
        content_types = webdav.pick_content_types(accepted_content_types, ["text/html"])
        assert content_types == ["text/html"]

        # Generate CalDAV/CardDAV URLs
        from urllib.parse import urlparse, urlunparse

        parsed = urlparse(self_url)

        # Determine if we're using HTTPS
        is_secure = parsed.scheme == "https"

        # Create URLs with different schemes, preserving the full path
        caldav_url = urlunparse(
            (
                "caldavs" if is_secure else "caldav",
                parsed.netloc,
                parsed.path,
                "",
                "",
                "",
            )
        )
        carddav_url = urlunparse(
            (
                "carddavs" if is_secure else "carddav",
                parsed.netloc,
                parsed.path,
                "",
                "",
                "",
            )
        )

        # Generate DAV×5 URL for QR code
        davx5_url = urlunparse(("davx5", parsed.netloc, parsed.path, "", "", ""))

        # Try to generate QR code if qrcode is available
        qr_code_data = None
        try:
            import qrcode
        except ImportError:
            logger.warning("qrcode package not installed; QR code generation disabled")
        else:
            import io
            import base64

            qr = qrcode.QRCode(version=1, box_size=10, border=4)
            qr.add_data(davx5_url)
            qr.make(fit=True)

            img = qr.make_image(fill_color="black", back_color="white")
            buffer = io.BytesIO()
            img.save(buffer, format="PNG")
            qr_code_data = base64.b64encode(buffer.getvalue()).decode()

        principals = self.backend.find_principals() if self.show_principals else []

        return render_jinja_page(
            "root.html",
            accepted_content_languages,
            principals=principals,
            self_url=self_url,
            caldav_url=caldav_url,
            carddav_url=carddav_url,
            davx5_url=davx5_url,
            qr_code_data=qr_code_data,
        )

    async def get_body(self):
        raise KeyError

    async def get_content_length(self):
        raise KeyError

    def get_content_type(self):
        return "text/html"

    def get_supported_locks(self):
        return []

    def get_active_locks(self):
        return []

    async def get_etag(self):
        h = hashlib.md5()
        for c in await self.get_body():
            h.update(c)
        return h.hexdigest()

    def get_last_modified(self):
        raise KeyError

    def get_content_language(self):
        return ["en-UK"]

    def get_member(self, name):
        return self.backend.get_resource("/" + name)

    def delete_member(self, name, etag=None, remote_user=None, requester=None):
        # This doesn't have any non-collection members.
        self.get_member("/" + name).destroy()

    def get_is_executable(self):
        return False

    def get_quota_used_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError

    def get_quota_available_bytes(self):
        # TODO(jelmer): Ask the store?
        raise KeyError


class Principal(webdav.Principal):
    def get_principal_url(self):
        return "."

    def get_principal_address(self):
        raise KeyError

    def get_calendar_home_set(self):
        return CALENDAR_HOME_SET

    def get_addressbook_home_set(self):
        return ADDRESSBOOK_HOME_SET

    def get_calendar_user_address_set(self):
        # TODO(jelmer): Make this configurable
        ret = []
        try:
            (fullname, email) = parseaddr(os.environ["EMAIL"])
        except KeyError:
            pass
        else:
            ret.append("mailto:" + email)
        return ret

    def set_infit_settings(self, settings):
        relpath = posixpath.join(self.relpath, ".infit")
        p = self.backend._map_to_file_path(relpath)
        with open(p, "w") as f:
            f.write(settings)

    def get_infit_settings(self):
        relpath = posixpath.join(self.relpath, ".infit")
        p = self.backend._map_to_file_path(relpath)
        if not os.path.exists(p):
            raise KeyError
        with open(p) as f:
            return f.read()

    def get_group_membership(self):
        """Get group membership URLs."""
        return []

    def get_calendar_user_type(self):
        # TODO(jelmer)
        return scheduling.CALENDAR_USER_TYPE_INDIVIDUAL

    def get_calendar_proxy_read_for(self):
        # TODO(jelmer)
        return []

    def get_calendar_proxy_write_for(self):
        # TODO(jelmer)
        return []

    def get_owner(self):
        return None

    def get_schedule_outbox_url(self):
        raise KeyError

    def get_schedule_inbox_url(self):
        # TODO(jelmer): make this configurable
        return "inbox"

    def get_creationdate(self):
        raise KeyError


class PrincipalBare(CollectionSetResource, Principal):
    """Principal user resource."""

    resource_types = [webdav.PRINCIPAL_RESOURCE_TYPE]

    @classmethod
    def create(cls, backend, relpath):
        p = super().create(backend, relpath)
        to_create = set()
        to_create.update(p.get_addressbook_home_set())
        to_create.update(p.get_calendar_home_set())
        for n in to_create:
            try:
                backend.create_collection(posixpath.join(relpath, n))
            except FileExistsError:
                pass
        return p

    async def render(
        self, self_url, accepted_content_types, accepted_content_languages
    ):
        content_types = webdav.pick_content_types(accepted_content_types, ["text/html"])
        assert content_types == ["text/html"]
        return await render_jinja_page(
            "principal.html",
            accepted_content_languages,
            principal=self,
            self_url=self_url,
        )

    def subcollections(self):
        # TODO(jelmer): Return members
        return []


class PrincipalCollection(Collection, Principal):
    """Principal user resource."""

    resource_types = webdav.Collection.resource_types + [webdav.PRINCIPAL_RESOURCE_TYPE]

    @classmethod
    def create(cls, backend, relpath):
        p = super().create(backend, relpath)
        p.store.set_type(STORE_TYPE_PRINCIPAL)
        to_create = set()
        to_create.update(p.get_addressbook_home_set())
        to_create.update(p.get_calendar_home_set())
        for n in to_create:
            try:
                backend.create_collection(posixpath.join(relpath, n))
            except FileExistsError:
                pass
        return p


class SingleUserFilesystemBackend(FilesystemBackend):
    def __init__(
        self,
        path,
        *,
        paranoid: bool = False,
        index_threshold: int | None = None,
        eager_indexing: bool = False,
        autocreate: bool = False,
        show_principals_on_root: bool = True,
    ) -> None:
        super().__init__(path)
        self._user_principals: set[str] = set()
        self.paranoid = paranoid
        self.index_threshold = index_threshold
        self.eager_indexing = eager_indexing
        self.autocreate = autocreate
        self.show_principals_on_root = show_principals_on_root
        self._open_store = functools.lru_cache(maxsize=16)(self._open_store_uncached)

    def _open_store_uncached(self, path: str):
        """Open a store from a filesystem path, uncached."""
        return open_store_from_path(
            path,
            double_check_indexes=self.paranoid,
            index_threshold=self.index_threshold,
            eager_indexing=self.eager_indexing,
        )

    def _mark_as_principal(self, path):
        self._user_principals.add(posixpath.normpath(path))

    def create_collection(self, relpath):
        p = self._map_to_file_path(relpath)
        store = TreeGitStore.create(p)
        self._open_store.cache_clear()
        return Collection(self, relpath, store)

    def create_principal(self, relpath, create_defaults=False):
        principal = PrincipalBare.create(self, relpath)
        self._mark_as_principal(relpath)
        if create_defaults:
            create_principal_defaults(self, principal)

    def find_principals(self):
        """List all of the principals on this server."""
        return self._user_principals

    def get_resource(self, relpath):
        relpath = posixpath.normpath(relpath)
        if not relpath.startswith("/"):
            raise ValueError("relpath %r should start with /")
        if relpath == "/":
            return RootPage(self, show_principals=self.show_principals_on_root)
        p = self._map_to_file_path(relpath)
        if p is None:
            return None
        if os.path.isdir(p):
            try:
                store = self._open_store(p)
            except NotStoreError:
                if relpath in self._user_principals:
                    return PrincipalBare(self, relpath)
                return CollectionSetResource(self, relpath)
            else:
                return {
                    STORE_TYPE_CALENDAR: CalendarCollection,
                    STORE_TYPE_ADDRESSBOOK: AddressbookCollection,
                    STORE_TYPE_PRINCIPAL: PrincipalCollection,
                    STORE_TYPE_SCHEDULE_INBOX: ScheduleInbox,
                    STORE_TYPE_SCHEDULE_OUTBOX: ScheduleOutbox,
                    STORE_TYPE_SUBSCRIPTION: SubscriptionCollection,
                    STORE_TYPE_OTHER: Collection,
                }[store.get_type()](self, relpath, store)
        else:
            (basepath, name) = os.path.split(relpath)
            assert name != "", f"path is {relpath!r}"
            store = self.get_resource(basepath)
            if store is None:
                return None
            if webdav.COLLECTION_RESOURCE_TYPE not in store.resource_types:
                return None
            try:
                return store.get_member(name)
            except KeyError:
                return None


class XandikosApp(webdav.WebDAVApp):
    """A wsgi App that provides a Xandikos web server."""

    def __init__(self, backend, current_user_principal, strict=True) -> None:
        super().__init__(backend, strict=strict)

        def get_current_user_principal(env):
            try:
                return current_user_principal % env
            except KeyError:
                return None

        self.register_properties(
            [
                webdav.ResourceTypeProperty(),
                webdav.CurrentUserPrincipalProperty(get_current_user_principal),
                webdav.PrincipalURLProperty(),
                webdav.DisplayNameProperty(),
                webdav.GetETagProperty(),
                webdav.GetContentTypeProperty(),
                webdav.GetContentLengthProperty(),
                webdav.GetContentLanguageProperty(),
                caldav.SourceProperty(),
                caldav.CalendarHomeSetProperty(),
                carddav.AddressbookHomeSetProperty(),
                caldav.CalendarDescriptionProperty(),
                caldav.CalendarColorProperty(),
                caldav.CalendarOrderProperty(),
                caldav.CreatedByProperty(),
                caldav.UpdatedByProperty(),
                caldav.SupportedCalendarComponentSetProperty(),
                carddav.AddressbookDescriptionProperty(),
                carddav.PrincipalAddressProperty(),
                webdav.AppleGetCTagProperty(),
                webdav.DAVGetCTagProperty(),
                carddav.SupportedAddressDataProperty(),
                webdav.SupportedReportSetProperty(self.reporters),
                sync.SyncTokenProperty(),
                caldav.SupportedCalendarDataProperty(),
                caldav.CalendarTimezoneProperty(),
                caldav.CalendarAvailabilityProperty(),
                caldav.MinDateTimeProperty(),
                caldav.MaxDateTimeProperty(),
                caldav.MaxResourceSizeProperty(),
                carddav.MaxResourceSizeProperty(),
                carddav.MaxImageSizeProperty(),
                access.CurrentUserPrivilegeSetProperty(),
                access.OwnerProperty(),
                webdav.CreationDateProperty(),
                webdav.SupportedLockProperty(),
                webdav.LockDiscoveryProperty(),
                infit.AddressbookColorProperty(),
                infit.SettingsProperty(),
                infit.HeaderValueProperty(),
                webdav.CommentProperty(),
                scheduling.CalendarUserAddressSetProperty(),
                scheduling.ScheduleInboxURLProperty(),
                scheduling.ScheduleOutboxURLProperty(),
                scheduling.CalendarUserTypeProperty(),
                scheduling.ScheduleTagProperty(),
                webdav.GetLastModifiedProperty(),
                timezones.TimezoneServiceSetProperty([]),
                webdav.AddMemberProperty(),
                caldav.ScheduleCalendarTransparencyProperty(),
                scheduling.ScheduleDefaultCalendarURLProperty(),
                caldav.MaxInstancesProperty(),
                caldav.MaxAttendeesPerInstanceProperty(),
                access.GroupMembershipProperty(),
                apache.ExecutableProperty(),
                caldav.CalendarProxyReadForProperty(),
                caldav.CalendarProxyWriteForProperty(),
                caldav.MaxAttachmentSizeProperty(),
                caldav.MaxAttachmentsPerResourceProperty(),
                caldav.ManagedAttachmentsServerURLProperty(),
                quota.QuotaAvailableBytesProperty(),
                quota.QuotaUsedBytesProperty(),
                webdav.RefreshRateProperty(),
                xmpp.XmppUriProperty(),
                xmpp.XmppServerProperty(),
                xmpp.XmppHeartbeatProperty(),
            ]
        )
        self.register_reporters(
            [
                caldav.CalendarMultiGetReporter(),
                caldav.CalendarQueryReporter(),
                carddav.AddressbookMultiGetReporter(),
                carddav.AddressbookQueryReporter(),
                webdav.ExpandPropertyReporter(),
                sync.SyncCollectionReporter(),
                caldav.FreeBusyQueryReporter(),
            ]
        )
        self.register_methods(
            [
                caldav.MkcalendarMethod(),
            ]
        )

    async def _handle_request(self, request, environ, start_response=None):
        if start_response and GIT_PATH in request.path.split(posixpath.sep):
            return self._handle_git_request(
                request,
                environ["ORIGINAL_ENVIRON"],
                takewhile(lambda x: x != GIT_PATH, request.path.split(posixpath.sep)),
                start_response,
            )
        else:
            return await super()._handle_request(request, environ)

    def _handle_git_request(self, request, environ, path, start_response):
        resource_path = posixpath.join("/", *path)
        resource = self.backend.get_resource(resource_path)
        if not isinstance(resource, StoreBasedCollection) or not isinstance(
            resource.store, GitStore
        ):
            return webdav._send_not_found(request)

        prefix = posixpath.join(resource_path, GIT_PATH)
        chain = make_wsgi_chain(DictBackend({prefix: resource.store.repo}), dumb=True)
        return chain(environ, start_response)


def create_principal_defaults(backend, principal):
    """Create default calendar and addressbook for a principal.

    Args:
      backend: Backend in which the principal exists.
      principal: Principal object
    """
    calendar_path = posixpath.join(
        principal.relpath, principal.get_calendar_home_set()[0], "calendar"
    )
    try:
        resource = backend.create_collection(calendar_path)
    except FileExistsError:
        pass
    else:
        resource.store.set_type(STORE_TYPE_CALENDAR)
        logger.info("Create calendar in %s.", resource.store.path)
    addressbook_path = posixpath.join(
        principal.relpath,
        principal.get_addressbook_home_set()[0],
        "addressbook",
    )
    try:
        resource = backend.create_collection(addressbook_path)
    except FileExistsError:
        pass
    else:
        resource.store.set_type(STORE_TYPE_ADDRESSBOOK)
        logger.info("Create addressbook in %s.", resource.store.path)
    calendar_path = posixpath.join(
        principal.relpath, principal.get_schedule_inbox_url()
    )
    try:
        resource = backend.create_collection(calendar_path)
    except FileExistsError:
        pass
    else:
        resource.store.set_type(STORE_TYPE_SCHEDULE_INBOX)
        logger.info("Create inbox in %s.", resource.store.path)


class RedirectDavHandler:
    def __init__(self, dav_root: str) -> None:
        self._dav_root = dav_root

    async def __call__(self, request):
        from aiohttp import web

        return web.HTTPFound(self._dav_root)


MDNS_NAME = "Xandikos CalDAV/CardDAV service"


def avahi_register(port: int, path: str):
    import avahi
    import dbus

    bus = dbus.SystemBus()
    server = dbus.Interface(
        bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER),
        avahi.DBUS_INTERFACE_SERVER,
    )
    group = dbus.Interface(
        bus.get_object(avahi.DBUS_NAME, server.EntryGroupNew()),
        avahi.DBUS_INTERFACE_ENTRY_GROUP,
    )

    for service in ["_carddav._tcp", "_caldav._tcp"]:
        try:
            group.AddService(
                avahi.IF_UNSPEC,
                avahi.PROTO_INET,
                0,
                MDNS_NAME,
                service,
                "",
                "",
                port,
                avahi.string_array_to_txt_array([f"path={path}"]),
            )
        except dbus.DBusException as e:
            logger.error("Error registering %s: %s", service, e)

    group.Commit()


def run_simple_server(
    directory: str,
    current_user_principal: str,
    autocreate: bool = False,
    defaults: bool = False,
    strict: bool = True,
    route_prefix: str = "/",
    listen_address: str | None = "::",
    port: int | None = 8080,
    socket_path: str | None = None,
) -> None:
    """Simple function to run a Xandikos server.

    This function is meant to be used by external code. We'll try our best
    not to break API compatibility.

    Args:
      directory: Directory to store data in ("/tmp/blah")
      current_user_principal: Name of current user principal ("/user")
      autocreate: Whether to create missing principals and collections
      defaults: Whether to create default calendar and addressbook collections
      strict: Whether to be strict in *DAV implementation. Set to False for
         buggy clients
      route_prefix: Route prefix under which to server ("/")
      listen_address: IP address to listen on (None to disable)
      port: TCP Port to listen on (None to disable)
      socket_path: Unix domain socket path to listen on (None to disable)
    """
    backend = SingleUserFilesystemBackend(directory)
    backend._mark_as_principal(current_user_principal)

    if autocreate or defaults:
        if not os.path.isdir(directory):
            os.makedirs(directory)
        backend.create_principal(current_user_principal, create_defaults=defaults)

    if not os.path.isdir(directory):
        logger.warning(
            "%r does not exist. Run xandikos with --autocreate?",
            directory,
        )
    if not backend.get_resource(current_user_principal):
        logger.warning(
            "default user principal %s does not exist. Run xandikos with --autocreate?",
            current_user_principal,
        )

    main_app = XandikosApp(
        backend,
        current_user_principal=current_user_principal,
        strict=strict,
    )

    async def xandikos_handler(request):
        return await main_app.aiohttp_handler(request, route_prefix)

    if socket_path:
        logger.info("Listening on unix domain socket %s", socket_path)
    if listen_address and port:
        logger.info("Listening on %s:%s", listen_address, port)

    from aiohttp import web

    app = web.Application()
    for path in WELLKNOWN_DAV_PATHS:
        app.router.add_route("*", path, RedirectDavHandler(route_prefix).__call__)

    if route_prefix.strip("/"):
        xandikos_app = web.Application()
        xandikos_app.router.add_route("*", "/{path_info:.*}", xandikos_handler)

        async def redirect_to_subprefix(request):
            return web.HTTPFound(route_prefix)

        app.router.add_route("*", "/", redirect_to_subprefix)
        app.add_subapp(route_prefix, xandikos_app)
    else:
        app.router.add_route("*", "/{path_info:.*}", xandikos_handler)

    web.run_app(app, port=port, host=listen_address, path=socket_path)


def add_parser(parser):
    import argparse

    access_group = parser.add_argument_group(title="Access Options")
    access_group.add_argument(
        "--no-detect-systemd",
        action="store_false",
        dest="detect_systemd",
        help="Disable systemd detection and socket activation.",
        default=systemd_imported,
    )
    access_group.add_argument(
        "-l",
        "--listen-address",
        dest="listen_address",
        default="localhost",
        help=(
            "Bind to this address. Pass in path for unix domain socket. [%(default)s]"
        ),
    )
    access_group.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        default=8080,
        help="Port to listen on. [%(default)s]",
    )
    access_group.add_argument(
        "--socket-mode",
        dest="socket_mode",
        default=None,
        help=(
            "File mode (permissions) for unix domain socket, "
            "in octal (e.g. 660). Only used when listening on a unix socket."
        ),
    )
    access_group.add_argument(
        "--socket-group",
        dest="socket_group",
        default=None,
        help=(
            "Group ownership for unix domain socket. "
            "Only used when listening on a unix socket."
        ),
    )
    access_group.add_argument(
        "--metrics-port",
        dest="metrics_port",
        default=None,
        help="Port to listen on for metrics. [%(default)s]",
    )
    access_group.add_argument(
        "--route-prefix",
        default="/",
        help=(
            "Path to Xandikos. "
            "(useful when Xandikos is behind a reverse proxy) "
            "[%(default)s]"
        ),
    )
    parser.add_argument(
        "-d",
        "--directory",
        dest="directory",
        default=None,
        required=True,
        help="Directory to serve from.",
    )
    parser.add_argument(
        "--current-user-principal",
        default="/user/",
        help="Path to current user principal. [%(default)s]",
    )
    parser.add_argument(
        "--autocreate",
        action="store_true",
        dest="autocreate",
        help="Automatically create necessary directories.",
    )
    parser.add_argument(
        "--defaults",
        action="store_true",
        dest="defaults",
        help=("Create initial calendar and address book. Implies --autocreate."),
    )
    parser.add_argument(
        "--dump-dav-xml",
        action="store_true",
        dest="dump_dav_xml",
        help="Print DAV XML request/responses.",
    )
    parser.add_argument(
        "--avahi", action="store_true", help="Announce services with avahi."
    )
    parser.add_argument(
        "--no-strict",
        action="store_false",
        dest="strict",
        help=("Enable workarounds for buggy CalDAV/CardDAV client implementations."),
        default=True,
    )
    parser.add_argument("--debug", action="store_true", help="Print debug messages")
    # Hidden arguments. These may change without notice in between releases,
    # and are generally just meant for developers.
    parser.add_argument("--paranoid", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--index-threshold", type=int, help=argparse.SUPPRESS)
    parser.add_argument(
        "--eager",
        action="store_true",
        help="Pre-populate indexes at startup for faster initial queries.",
    )


async def main(options, parser):
    if options.dump_dav_xml:
        # TODO(jelmer): Find a way to propagate this without abusing
        # os.environ.
        os.environ["XANDIKOS_DUMP_DAV_XML"] = "1"

    if not options.route_prefix.endswith("/"):
        options.route_prefix += "/"

    if options.debug:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.INFO

    logging.basicConfig(level=loglevel, format="%(message)s")

    backend = SingleUserFilesystemBackend(
        os.path.abspath(options.directory),
        paranoid=options.paranoid,
        index_threshold=options.index_threshold,
        eager_indexing=options.eager,
    )
    backend._mark_as_principal(options.current_user_principal)

    if options.autocreate or options.defaults:
        if not os.path.isdir(options.directory):
            os.makedirs(options.directory)
        backend.create_principal(
            options.current_user_principal, create_defaults=options.defaults
        )

    if not os.path.isdir(options.directory):
        logger.warning(
            "%r does not exist. Run xandikos with --autocreate?",
            options.directory,
        )
    if not backend.get_resource(options.current_user_principal):
        logger.warning(
            "default user principal %s does not exist. Run xandikos with --autocreate?",
            options.current_user_principal,
        )

    from .__main__ import _get_package_versions

    version_str = ", ".join(f"{pkg} {ver}" for pkg, ver in _get_package_versions())
    logger.info("%s", version_str)

    main_app = XandikosApp(
        backend,
        current_user_principal=options.current_user_principal,
        strict=options.strict,
    )

    async def xandikos_handler(request):
        return await main_app.aiohttp_handler(request, options.route_prefix)

    if options.detect_systemd and not systemd_imported:
        parser.error("systemd detection requested, but unable to find systemd_python")

    if options.detect_systemd and systemd.daemon.booted():
        listen_socks = get_systemd_listen_sockets()
        socket_path = None
        listen_address = None
        listen_port = None
        logger.info("Receiving file descriptors from systemd socket activation")
    elif "/" in options.listen_address:
        socket_path = options.listen_address
        listen_address = None
        listen_port = None  # otherwise aiohttp also listens on default host
        listen_socks = []
        logger.info("Listening on unix domain socket %s", socket_path)
    else:
        listen_address = options.listen_address
        listen_port = options.port
        socket_path = None
        listen_socks = []
        logger.info("Listening on %s:%s", listen_address, options.port)

    from aiohttp import web

    if options.metrics_port == options.port:
        parser.error("Metrics port cannot be the same as the main port")

    app = web.Application()
    if options.metrics_port is not None:
        metrics_app = web.Application()
        try:
            from aiohttp_openmetrics import metrics, metrics_middleware
        except ModuleNotFoundError:
            logger.warning(
                "aiohttp-openmetrics not found; /metrics will not be available."
            )
        else:
            app.middlewares.insert(0, metrics_middleware)
            metrics_app.router.add_get("/metrics", metrics, name="metrics")

        # For now, just always claim everything is okay.
        metrics_app.router.add_get("/health", lambda r: web.Response(text="ok"))
    else:
        metrics_app = None

    for path in WELLKNOWN_DAV_PATHS:
        app.router.add_route(
            "*", path, RedirectDavHandler(options.route_prefix).__call__
        )

    if options.route_prefix.strip("/"):
        xandikos_app = web.Application()
        xandikos_app.router.add_route("*", "/{path_info:.*}", xandikos_handler)

        async def redirect_to_subprefix(request):
            return web.HTTPFound(options.route_prefix)

        app.router.add_route("*", "/", redirect_to_subprefix)
        app.add_subapp(options.route_prefix, xandikos_app)
    else:
        app.router.add_route("*", "/{path_info:.*}", xandikos_handler)

    if options.avahi:
        try:
            import avahi  # noqa: F401
            import dbus  # noqa: F401
        except ImportError:
            logger.error(
                "Please install python-avahi and python-dbus for avahi support."
            )
        else:
            avahi_register(options.port, options.route_prefix)

    runner = web.AppRunner(app)
    await runner.setup()
    sites = []
    if metrics_app:
        metrics_runner = web.AppRunner(metrics_app)
        await metrics_runner.setup()
        # TODO(jelmer): Allow different metrics listen address?
        sites.append(web.TCPSite(metrics_runner, listen_address, options.metrics_port))
    # Use systemd sockets first and only if not present use the socket path or
    # address from --listen-address.
    if listen_socks:
        sites.extend([web.SockSite(runner, sock) for sock in listen_socks])
    elif socket_path:
        sites.append(web.UnixSite(runner, socket_path))
    else:
        sites.append(web.TCPSite(runner, listen_address, listen_port))

    import signal

    # Set up graceful shutdown handling
    loop = asyncio.get_running_loop()
    shutdown_event = asyncio.Event()

    def signal_handler(signum, frame):
        logger.info("Received signal %s, shutting down gracefully...", signum)
        # Use call_soon_threadsafe to safely set the event from signal handler
        loop.call_soon_threadsafe(shutdown_event.set)

    # Register signal handlers for graceful shutdown
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)

    for site in sites:
        await site.start()

    # Set socket group ownership after the socket is created
    if socket_path and options.socket_group is not None:
        import grp

        try:
            gid = grp.getgrnam(options.socket_group).gr_gid
            os.chown(socket_path, -1, gid)
            logger.info("Set socket group to %s", options.socket_group)
        except KeyError:
            parser.error(f"Unknown group: {options.socket_group}")
        except OSError as e:
            logger.error("Failed to set socket group: %s", e)

    # Set socket permissions after the socket is created
    if socket_path and options.socket_mode is not None:
        try:
            mode = int(options.socket_mode, 8)
            os.chmod(socket_path, mode)
            logger.info("Set socket permissions to %s", options.socket_mode)
        except ValueError:
            parser.error(f"Invalid socket mode: {options.socket_mode}")
        except OSError as e:
            logger.error("Failed to set socket permissions: %s", e)

    # Wait for shutdown signal
    try:
        await shutdown_event.wait()
    except KeyboardInterrupt:
        logger.info("Received KeyboardInterrupt, shutting down gracefully...")

    # Cleanup: stop all sites and runners
    logger.info("Stopping web servers...")
    for site in sites:
        await site.stop()

    await runner.cleanup()
    if metrics_app:
        await metrics_runner.cleanup()

    logger.info("Shutdown complete.")


if __name__ == "__main__":
    import sys

    import argparse

    parser = argparse.ArgumentParser(usage="%(prog)s [options]")
    add_parser(parser)
    args = parser.parse_args(sys.argv[1:])

    sys.exit(asyncio.run(main(args, parser)))



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""CardDAV support.

https://tools.ietf.org/html/rfc6352
"""

import itertools

from . import collation as _mod_collation
from . import davcommon, webdav

ET = webdav.ET

WELLKNOWN_CARDDAV_PATH = "/.well-known/carddav"

NAMESPACE = "urn:ietf:params:xml:ns:carddav"
ADDRESSBOOK_RESOURCE_TYPE = "{%s}addressbook" % NAMESPACE

# Feature to advertise presence of CardDAV support
FEATURE = "addressbook"


class AddressbookHomeSetProperty(webdav.Property):
    """addressbook-home-set property.

    See https://tools.ietf.org/html/rfc6352, section 7.1.1
    """

    name = "{%s}addressbook-home-set" % NAMESPACE
    resource_type = "{DAV:}principal"
    in_allprops = False
    live = True

    async def get_value(self, base_href, resource, el, environ):
        for href in resource.get_addressbook_home_set():
            href = webdav.ensure_trailing_slash(href)
            el.append(webdav.create_href(href, base_href))


class AddressDataProperty(davcommon.SubbedProperty):
    """address-data property.

    See https://tools.ietf.org/html/rfc6352, section 10.4

    Note that this is not technically a DAV property, and
    it is thus not registered in the regular webdav server.
    """

    name = "{%s}address-data" % NAMESPACE

    def supported_on(self, resource):
        try:
            return resource.get_content_type() == "text/vcard"
        except KeyError:
            # Resource doesn't have a content type
            return False

    async def get_value_ext(self, href, resource, el, environ, requested):
        # TODO(jelmer): Support subproperties
        # UTF-8 encoding is required by RFC 6350 (vCard format)
        el.text = b"".join(await resource.get_body()).decode("utf-8")


class AddressbookDescriptionProperty(webdav.Property):
    """Provides calendar-description property.

    https://tools.ietf.org/html/rfc6352, section 6.2.1
    """

    name = "{%s}addressbook-description" % NAMESPACE
    resource_type = ADDRESSBOOK_RESOURCE_TYPE

    async def get_value(self, href, resource, el, environ):
        el.text = resource.get_addressbook_description()

    async def set_value(self, href, resource, el):
        resource.set_addressbook_description(el.text)


class AddressbookMultiGetReporter(davcommon.MultiGetReporter):
    name = "{%s}addressbook-multiget" % NAMESPACE
    resource_type = ADDRESSBOOK_RESOURCE_TYPE
    data_property = AddressDataProperty()

    async def report(
        self,
        environ,
        body,
        resources_by_hrefs,
        properties,
        base_href,
        resource,
        depth,
        strict,
    ):
        # RFC 6352 Section 8.7 (CardDAV addressbook-multiget) specifies:
        #   "The request MUST include a Depth: 0 header on the request."
        #
        # This is a client requirement, and the RFC doesn't explicitly mandate
        # that servers MUST reject requests with other Depth values. However,
        # in strict mode, we enforce this requirement to ensure full RFC
        # compliance and catch misbehaving clients.
        #
        # In non-strict mode, we accept any Depth value for compatibility with
        # existing clients that may send non-zero Depth headers.
        if strict and depth != "0":
            raise webdav.BadRequestError(
                f"{self.name} requires Depth: 0 (RFC 6352 Section 8.7), "
                f"got Depth: {depth}"
            )
        return await super().report(
            environ,
            body,
            resources_by_hrefs,
            properties,
            base_href,
            resource,
            depth,
            strict,
        )


class Addressbook(webdav.Collection):
    resource_types = webdav.Collection.resource_types + [ADDRESSBOOK_RESOURCE_TYPE]

    def get_addressbook_description(self) -> str:
        raise NotImplementedError(self.get_addressbook_description)

    def set_addressbook_description(self, description: str) -> None:
        raise NotImplementedError(self.set_addressbook_description)

    def get_addressbook_color(self) -> str:
        raise NotImplementedError(self.get_addressbook_color)

    def set_addressbook_color(self, color: str) -> None:
        raise NotImplementedError(self.set_addressbook_color)

    def get_supported_address_data_types(self):
        """Get list of supported data types.

        Returns: List of tuples with content type and version
        """
        raise NotImplementedError(self.get_supported_address_data_types)

    def get_max_resource_size(self) -> int:
        """Get maximum object size this address book will store (in bytes).

        Absence indicates no maximum.
        """
        raise NotImplementedError(self.get_max_resource_size)

    def get_max_image_size(self) -> int:
        """Get maximum image size this address book will store (in bytes).

        Absence indicates no maximum.
        """
        raise NotImplementedError(self.get_max_image_size)


class PrincipalExtensions:
    """Extensions to webdav.Principal."""

    def get_addressbook_home_set(self) -> set[str]:
        """Return set of addressbook home URLs.

        Returns: set of URLs
        """
        raise NotImplementedError(self.get_addressbook_home_set)

    def get_principal_address(self) -> str:
        """Return URL to principal address vCard."""
        raise NotImplementedError(self.get_principal_address)


class PrincipalAddressProperty(webdav.Property):
    """Provides the principal-address property.

    https://tools.ietf.org/html/rfc6352, section 7.1.2
    """

    name = "{%s}principal-address" % NAMESPACE
    resource_type = "{DAV:}principal"
    in_allprops = False

    async def get_value(self, href, resource, el, environ):
        el.append(webdav.create_href(resource.get_principal_address(), href))


class SupportedAddressDataProperty(webdav.Property):
    """Provides the supported-address-data property.

    https://tools.ietf.org/html/rfc6352, section 6.2.2
    """

    name = "{%s}supported-address-data" % NAMESPACE
    resource_type = ADDRESSBOOK_RESOURCE_TYPE
    in_allprops = False
    live = True

    async def get_value(self, href, resource, el, environ):
        for (
            content_type,
            version,
        ) in resource.get_supported_address_data_types():
            subel = ET.SubElement(el, "{%s}content-type" % NAMESPACE)
            subel.set("content-type", content_type)
            subel.set("version", version)


class MaxResourceSizeProperty(webdav.Property):
    """Provides the max-resource-size property.

    See https://tools.ietf.org/html/rfc6352, section 6.2.3.
    """

    name = "{%s}max-resource-size" % NAMESPACE
    resource_type = ADDRESSBOOK_RESOURCE_TYPE
    in_allprops = False
    live = True

    async def get_value(self, href, resource, el, environ):
        el.text = str(resource.get_max_resource_size())


class MaxImageSizeProperty(webdav.Property):
    """Provides the max-image-size property.

    This seems to be a carddav extension used by iOS and caldavzap.
    """

    name = "{%s}max-image-size" % NAMESPACE
    resource_type = ADDRESSBOOK_RESOURCE_TYPE
    in_allprops = False
    live = True

    async def get_value(self, href, resource, el, environ):
        el.text = str(resource.get_max_image_size())


async def addressbook_from_resource(resource):
    try:
        if resource.get_content_type() != "text/vcard":
            return None
    except KeyError:
        return None
    file = await resource.get_file()
    return file.addressbook.contents


def apply_text_match(el: ET.Element, value: str) -> bool:
    collation = el.get("collation", "i;ascii-casemap")
    negate_condition = el.get("negate-condition", "no")
    match_type = el.get("match-type", "contains")
    matches = _mod_collation.collations[collation](value, el.text or "", match_type)

    if negate_condition == "yes":
        return not matches
    else:
        return matches


def apply_param_filter(el, prop):
    name = el.get("name")
    if len(el) == 1 and el[0].tag == "{urn:ietf:params:xml:ns:carddav}is-not-defined":
        return name not in prop.params

    try:
        value = prop.params[name]
    except KeyError:
        return False

    for subel in el:
        if subel.tag == "{urn:ietf:params:xml:ns:carddav}text-match":
            if not apply_text_match(subel, value):
                return False
        else:
            raise AssertionError("unknown tag %r in param-filter", subel.tag)
    return True


def apply_prop_filter(el, ab):
    name = el.get("name").lower()
    # From https://tools.ietf.org/html/rfc6352
    # A CARDDAV:prop-filter is said to match if:

    # The CARDDAV:prop-filter XML element contains a CARDDAV:is-not-defined XML
    # element and no property of the type specified by the "name" attribute
    # exists in the enclosing calendar component;
    if len(el) == 1 and el[0].tag == "{urn:ietf:params:xml:ns:carddav}is-not-defined":
        return name not in ab

    try:
        prop = ab[name]
    except KeyError:
        return False

    for prop_el in prop:
        matched = True
        for subel in el:
            if subel.tag == "{urn:ietf:params:xml:ns:carddav}text-match":
                if not apply_text_match(subel, str(prop_el)):
                    matched = False
                    break
            elif subel.tag == "{urn:ietf:params:xml:ns:carddav}param-filter":
                if not apply_param_filter(subel, prop_el):
                    matched = False
                    break
        if matched:
            return True
    return False


def parse_filter(filter_el, cls):
    """Parse a CardDAV filter element and build a filter object."""
    if filter_el is None:
        return cls

    test_name = filter_el.get("test", "anyof")
    cls.test = {"allof": all, "anyof": any}[test_name]

    for prop_el in filter_el:
        if prop_el.tag == "{urn:ietf:params:xml:ns:carddav}prop-filter":
            parse_prop_filter(prop_el, cls)
        else:
            raise AssertionError(f"unknown filter tag {prop_el.tag!r}")

    return cls


def parse_prop_filter(prop_el, filter_obj):
    """Parse a prop-filter element and add it to the filter."""
    name = prop_el.get("name")
    text_match = None
    param_filters = []
    is_not_defined = False

    for subel in prop_el:
        if subel.tag == "{urn:ietf:params:xml:ns:carddav}is-not-defined":
            is_not_defined = True
        elif subel.tag == "{urn:ietf:params:xml:ns:carddav}text-match":
            text_match = {
                "text": subel.text or "",
                "collation": subel.get("collation", "i;unicode-casemap"),
                "negate_condition": subel.get("negate-condition", "no") == "yes",
                "match_type": subel.get("match-type", "contains"),
            }
        elif subel.tag == "{urn:ietf:params:xml:ns:carddav}param-filter":
            # param-filter is handled by apply_param_filter() during query execution
            pass

    filter_obj.add_prop_filter(name, text_match, param_filters, is_not_defined)


async def apply_filter(el, resource):
    """Compile a filter element into a Python function."""
    if el is None or not list(el):
        # Empty filter, let's not bother parsing
        return lambda x: True
    ab = await addressbook_from_resource(resource)
    if ab is None:
        return False
    test_name = el.get("test", "anyof")
    test = {"allof": all, "anyof": any}[test_name]
    return test(apply_prop_filter(subel, ab) for subel in el)


class AddressbookQueryReporter(webdav.Reporter):
    name = "{%s}addressbook-query" % NAMESPACE
    resource_type = ADDRESSBOOK_RESOURCE_TYPE
    data_property = AddressDataProperty()

    @webdav.multistatus
    async def report(
        self,
        environ,
        body,
        resources_by_hrefs,
        properties,
        base_href,
        base_resource,
        depth,
        strict,
    ):
        requested = None
        filter_el = None
        limit = None
        for el in body:
            if el.tag in ("{DAV:}prop", "{DAV:}allprop", "{DAV:}propname"):
                requested = el
            elif el.tag == ("{%s}filter" % NAMESPACE):
                filter_el = el
            elif el.tag == ("{%s}limit" % NAMESPACE):
                limit = el
            else:
                webdav.nonfatal_bad_request(
                    f"Unknown tag {el.tag} in report {self.name}", strict
                )
        if requested is None:
            # The CardDAV RFC says that behaviour mimics that of PROPFIND,
            # and the WebDAV RFC says that no body implies {DAV}allprop
            # This isn't exactly an empty body, but close enough.
            requested = ET.Element("{DAV:}allprop")
        if limit is not None:
            try:
                [nresults_el] = list(limit)
            except ValueError:
                webdav.nonfatal_bad_request(
                    "Invalid number of subelements in limit", strict
                )
                nresults = None
            else:
                try:
                    nresults = int(nresults_el.text)
                except ValueError:
                    webdav.nonfatal_bad_request("nresults not a number", strict)
                    nresults = None
        else:
            nresults = None

        def filter_fn(cls):
            return parse_filter(filter_el, cls())

        def members(collection):
            return itertools.chain(
                collection.addressbook_query(filter_fn),
                collection.subcollections(),
            )

        i = 0
        async for href, resource in webdav.traverse_resource(
            base_resource, base_href, depth, members=members
        ):
            # Ideally traverse_resource would only return the right things.
            if getattr(resource, "content_type", None) == "text/vcard":
                if nresults is not None and i >= nresults:
                    break
                propstat = davcommon.get_properties_with_data(
                    self.data_property,
                    href,
                    resource,
                    properties,
                    environ,
                    requested,
                )
                yield webdav.Status(
                    href, "200 OK", propstat=[s async for s in propstat]
                )
                i += 1



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Collations."""

from collections.abc import Callable


class UnknownCollation(Exception):
    def __init__(self, collation: str) -> None:
        super().__init__(f"Collation {collation!r} is not supported")
        self.collation = collation


def _match(a, b, k):
    if k == "equals":
        return a == b
    elif k == "contains":
        return b in a
    elif k == "starts-with":
        return a.startswith(b)
    elif k == "ends-with":
        return a.endswith(b)
    else:
        raise NotImplementedError


collations: dict[str, Callable[[str, str, str], bool]] = {
    "i;ascii-casemap": lambda a, b, k: _match(
        a.encode("ascii").upper(), b.encode("ascii").upper(), k
    ),
    "i;octet": lambda a, b, k: _match(a, b, k),
    # TODO(jelmer): Follow all rules as specified in
    # https://datatracker.ietf.org/doc/html/rfc5051
    "i;unicode-casemap": lambda a, b, k: _match(
        a.encode("utf-8", "surrogateescape").upper(),
        b.encode("utf-8", "surrogateescape").upper(),
        k,
    ),
}


def get_collation(name: str) -> Callable[[str, str, str], bool]:
    """Get a collation by name.

    Args:
      name: Collation name
    Raises:
      UnknownCollation: If the collation is not supported
    """
    try:
        return collations[name]
    except KeyError as exc:
        raise UnknownCollation(name) from exc



# Xandikos
# Copyright (C) 2025 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Performance tests for large collections (RFC 4918, RFC 4791, RFC 6352).

These tests ensure that Xandikos performs adequately with large collections
of calendar events and contacts.
"""

import time
import unittest
from datetime import datetime, timedelta, timezone

from xandikos.icalendar import CalendarFilter, ICalendarFile
from xandikos.vcard import CardDAVFilter, VCardFile


class LargeCollectionPerformanceTests(unittest.TestCase):
    """Performance tests for large collections."""

    def setUp(self):
        """Set up large test collections."""
        # Create a large CalDAV calendar (500 events)
        self.large_calendar_events = []
        base_date = datetime(2025, 1, 1, 10, 0, 0, tzinfo=timezone.utc)

        for i in range(500):
            event_date = base_date + timedelta(days=i)
            event = ICalendarFile(
                [
                    b"BEGIN:VCALENDAR\r\n",
                    b"VERSION:2.0\r\n",
                    b"PRODID:-//Test//Test//EN\r\n",
                    b"BEGIN:VEVENT\r\n",
                    f"UID:event-{i}@test.example.com\r\n".encode(),
                    f"DTSTART:{event_date.strftime('%Y%m%dT%H%M%SZ')}\r\n".encode(),
                    f"DTEND:{(event_date + timedelta(hours=1)).strftime('%Y%m%dT%H%M%SZ')}\r\n".encode(),
                    f"SUMMARY:Event {i}\r\n".encode(),
                    b"END:VEVENT\r\n",
                    b"END:VCALENDAR\r\n",
                ],
                "text/calendar",
            )
            self.large_calendar_events.append((f"event-{i}.ics", event))

        # Create a large CardDAV addressbook (500 contacts)
        self.large_addressbook_contacts = []
        for i in range(500):
            vcard = VCardFile(
                [
                    b"BEGIN:VCARD\r\n",
                    b"VERSION:3.0\r\n",
                    f"FN:Contact {i}\r\n".encode(),
                    f"N:Contact{i};Test;;;\r\n".encode(),
                    f"EMAIL:contact{i}@example.com\r\n".encode(),
                    f"TEL:+1-555-{i:04d}\r\n".encode(),
                    b"END:VCARD\r\n",
                ],
                "text/vcard",
            )
            self.large_addressbook_contacts.append((f"contact-{i}.vcf", vcard))

    def test_caldav_filter_performance_no_match(self):
        """Test CalDAV filter performance on large calendar (no matches).

        This tests worst-case performance where the filter must check
        all events but finds no matches.
        """
        filter_obj = CalendarFilter(timezone.utc)
        comp_filter = filter_obj.filter_subcomponent("VCALENDAR").filter_subcomponent(
            "VEVENT"
        )
        comp_filter.filter_property("SUMMARY").filter_text_match(
            "NonexistentEvent", collation="i;unicode-casemap"
        )

        start_time = time.perf_counter()
        matches = [
            name
            for name, event in self.large_calendar_events
            if filter_obj.check(name, event)
        ]
        elapsed = time.perf_counter() - start_time

        self.assertEqual(0, len(matches))
        # Should complete in under 5 seconds for 500 events
        self.assertLess(elapsed, 5.0, f"Filter took {elapsed:.2f}s, expected < 5s")

    def test_caldav_filter_performance_some_matches(self):
        """Test CalDAV filter performance with partial matches.

        This tests typical performance where some events match the filter.
        """
        filter_obj = CalendarFilter(timezone.utc)
        comp_filter = filter_obj.filter_subcomponent("VCALENDAR").filter_subcomponent(
            "VEVENT"
        )
        # Match events with "Event 1" in summary (Event 1, Event 10-19, Event 100-199)
        comp_filter.filter_property("SUMMARY").filter_text_match(
            "Event 1", collation="i;unicode-casemap"
        )

        start_time = time.perf_counter()
        matches = [
            name
            for name, event in self.large_calendar_events
            if filter_obj.check(name, event)
        ]
        elapsed = time.perf_counter() - start_time

        # Should find "Event 1", "Event 10"-"Event 19", "Event 100"-"Event 199"
        self.assertGreater(len(matches), 100)
        self.assertLess(elapsed, 5.0, f"Filter took {elapsed:.2f}s, expected < 5s")

    def test_caldav_time_range_filter_performance(self):
        """Test CalDAV time-range filter performance.

        This tests performance of time-range queries, which are common
        in calendar applications.
        """
        filter_obj = CalendarFilter(timezone.utc)
        comp_filter = filter_obj.filter_subcomponent("VCALENDAR").filter_subcomponent(
            "VEVENT"
        )
        # Query for events in January 2025 (first 31 events)
        start = datetime(2025, 1, 1, tzinfo=timezone.utc)
        end = datetime(2025, 2, 1, tzinfo=timezone.utc)
        comp_filter.filter_time_range(start, end)

        start_time = time.perf_counter()
        matches = [
            name
            for name, event in self.large_calendar_events
            if filter_obj.check(name, event)
        ]
        elapsed = time.perf_counter() - start_time

        # Should find 31 events (January has 31 days)
        self.assertEqual(31, len(matches))
        self.assertLess(elapsed, 5.0, f"Filter took {elapsed:.2f}s, expected < 5s")

    def test_carddav_filter_performance_no_match(self):
        """Test CardDAV filter performance on large addressbook (no matches).

        This tests worst-case performance where the filter must check
        all contacts but finds no matches.
        """
        filter_obj = CardDAVFilter()
        filter_obj.add_property_filter("FN").add_text_match(
            "NonexistentContact", match_type="contains"
        )

        start_time = time.perf_counter()
        matches = [
            name
            for name, vcard in self.large_addressbook_contacts
            if filter_obj.check(name, vcard)
        ]
        elapsed = time.perf_counter() - start_time

        self.assertEqual(0, len(matches))
        # Should complete in under 2 seconds for 500 contacts
        self.assertLess(elapsed, 2.0, f"Filter took {elapsed:.2f}s, expected < 2s")

    def test_carddav_filter_performance_some_matches(self):
        """Test CardDAV filter performance with partial matches.

        This tests typical performance where some contacts match the filter.
        """
        filter_obj = CardDAVFilter()
        # Match contacts with "Contact 1" in FN
        filter_obj.add_property_filter("FN").add_text_match(
            "Contact 1", match_type="contains"
        )

        start_time = time.perf_counter()
        matches = [
            name
            for name, vcard in self.large_addressbook_contacts
            if filter_obj.check(name, vcard)
        ]
        elapsed = time.perf_counter() - start_time

        # Should find "Contact 1", "Contact 10"-"Contact 19", "Contact 100"-"Contact 199"
        self.assertGreater(len(matches), 100)
        self.assertLess(elapsed, 2.0, f"Filter took {elapsed:.2f}s, expected < 2s")

    def test_carddav_email_filter_performance(self):
        """Test CardDAV email filter performance.

        This tests performance of email-based queries, which are common
        in addressbook applications.
        """
        filter_obj = CardDAVFilter()
        # Match contacts with specific email domain
        filter_obj.add_property_filter("EMAIL").add_text_match(
            "example.com", match_type="contains"
        )

        start_time = time.perf_counter()
        matches = [
            name
            for name, vcard in self.large_addressbook_contacts
            if filter_obj.check(name, vcard)
        ]
        elapsed = time.perf_counter() - start_time

        # All contacts should match (all have @example.com)
        self.assertEqual(500, len(matches))
        self.assertLess(elapsed, 2.0, f"Filter took {elapsed:.2f}s, expected < 2s")

    def test_caldav_complex_filter_performance(self):
        """Test CalDAV complex filter with multiple conditions.

        This tests performance of complex queries with multiple filter
        conditions combined.
        """
        filter_obj = CalendarFilter(timezone.utc)
        filter_obj.test = all  # AND logic
        comp_filter = filter_obj.filter_subcomponent("VCALENDAR").filter_subcomponent(
            "VEVENT"
        )

        # Multiple conditions: text match + time range
        comp_filter.filter_property("SUMMARY").filter_text_match(
            "Event", collation="i;unicode-casemap"
        )
        start = datetime(2025, 1, 1, tzinfo=timezone.utc)
        end = datetime(2025, 2, 1, tzinfo=timezone.utc)
        comp_filter.filter_time_range(start, end)

        start_time = time.perf_counter()
        matches = [
            name
            for name, event in self.large_calendar_events
            if filter_obj.check(name, event)
        ]
        elapsed = time.perf_counter() - start_time

        # Should find 31 events in January
        self.assertEqual(31, len(matches))
        self.assertLess(elapsed, 5.0, f"Filter took {elapsed:.2f}s, expected < 5s")

    def test_carddav_complex_filter_performance(self):
        """Test CardDAV complex filter with multiple conditions.

        This tests performance of complex queries with multiple filter
        conditions combined.
        """
        filter_obj = CardDAVFilter()
        filter_obj.test = all  # AND logic
        filter_obj.add_property_filter("EMAIL")
        filter_obj.add_property_filter("TEL")
        filter_obj.add_property_filter("FN").add_text_match(
            "Contact 1", match_type="contains"
        )

        start_time = time.perf_counter()
        matches = [
            name
            for name, vcard in self.large_addressbook_contacts
            if filter_obj.check(name, vcard)
        ]
        elapsed = time.perf_counter() - start_time

        # Should find contacts matching all conditions
        self.assertGreater(len(matches), 100)
        self.assertLess(elapsed, 2.0, f"Filter took {elapsed:.2f}s, expected < 2s")


if __name__ == "__main__":
    unittest.main()



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""WSGI wrapper for xandikos."""

from logging import getLogger
import os

from .web import XandikosApp, SingleUserFilesystemBackend

logger = getLogger("xandikos")

create_defaults = False

autocreate_str = os.getenv("AUTOCREATE")
if autocreate_str == "defaults":
    logger.warning("Creating default collections.")
    create_defaults = True
    autocreate = True
elif autocreate_str in ("empty", "yes"):
    autocreate = True
elif autocreate_str in (None, "no"):
    autocreate = False
else:
    logger.warning("Unknown value for AUTOCREATE: %r", autocreate_str)
    autocreate = False

eager_indexing = os.getenv("EAGER", "").lower() in ("true", "1", "yes")
backend = SingleUserFilesystemBackend(
    path=os.environ["XANDIKOSPATH"],
    eager_indexing=eager_indexing,
)
if not os.path.isdir(backend.path):
    if autocreate:
        os.makedirs(os.environ["XANDIKOSPATH"])
    else:
        logger.warning("%r does not exist.", backend.path)

current_user_principal = os.environ.get("CURRENT_USER_PRINCIPAL", "/user/")
if not backend.get_resource(current_user_principal):
    if autocreate:
        backend.create_principal(
            current_user_principal, create_defaults=create_defaults
        )
    else:
        logger.warning(
            "default user principal '%s' does not exist. "
            "Create directory %s or set AUTOCREATE variable?",
            current_user_principal,
            backend._map_to_file_path(current_user_principal),
        )

backend._mark_as_principal(current_user_principal)
app = XandikosApp(backend, current_user_principal)



# Xandikos
# Copyright (C) 2025 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.

"""Tests for xandikos.access (RFC 3744 WebDAV ACL)."""

import asyncio
import unittest

from xandikos import access, webdav
from xandikos.webdav import ET


class CurrentUserPrivilegeSetPropertyTests(unittest.TestCase):
    """Tests for CurrentUserPrivilegeSetProperty (RFC 3744 Section 3.7)."""

    def test_property_name(self):
        """Test current-user-privilege-set property name."""
        prop = access.CurrentUserPrivilegeSetProperty()
        self.assertEqual(prop.name, "{DAV:}current-user-privilege-set")

    def test_property_attributes(self):
        """Test current-user-privilege-set property attributes.

        RFC 3744 Section 3.7: This is a protected property that
        identifies the privileges granted to the current user.
        """
        prop = access.CurrentUserPrivilegeSetProperty()
        self.assertFalse(prop.in_allprops)
        self.assertTrue(prop.live)

    def test_get_value(self):
        """Test current-user-privilege-set get_value method.

        RFC 3744 Section 3.7: Contains the privileges currently
        granted to the authenticated user.
        """

        async def run_test():
            prop = access.CurrentUserPrivilegeSetProperty()

            class MockResource:
                pass

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/test", resource, el, {})

            # Should have a privilege element
            privileges = el.findall("{DAV:}privilege")
            self.assertEqual(len(privileges), 1)

            # Should contain DAV:all privilege (current implementation)
            all_priv = privileges[0].find("{DAV:}all")
            self.assertIsNotNone(all_priv)

        asyncio.run(run_test())

    def test_get_value_xml_structure(self):
        """Test that current-user-privilege-set generates correct XML structure."""

        async def run_test():
            prop = access.CurrentUserPrivilegeSetProperty()

            class MockResource:
                pass

            resource = MockResource()
            el = ET.Element("{DAV:}current-user-privilege-set")

            await prop.get_value("/test", resource, el, {})

            # Verify XML structure matches RFC 3744 format
            xml_str = ET.tostring(el, encoding="unicode")
            self.assertIn("<ns0:privilege", xml_str)
            self.assertIn("<ns0:all", xml_str)

        asyncio.run(run_test())


class OwnerPropertyTests(unittest.TestCase):
    """Tests for OwnerProperty (RFC 3744 Section 5.1)."""

    def test_property_name(self):
        """Test owner property name."""
        prop = access.OwnerProperty()
        self.assertEqual(prop.name, "{DAV:}owner")

    def test_property_attributes(self):
        """Test owner property attributes.

        RFC 3744 Section 5.1: Identifies a resource as the owner
        of the resource.
        """
        prop = access.OwnerProperty()
        self.assertFalse(prop.in_allprops)
        self.assertTrue(prop.live)

    def test_get_value_with_owner(self):
        """Test owner property when resource has an owner.

        RFC 3744 Section 5.1: The owner is typically the principal
        who created the resource.
        """

        async def run_test():
            prop = access.OwnerProperty()

            class MockResource:
                def get_owner(self):
                    return "/principals/user1/"

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/resource", resource, el, {})

            # Should have an href element with the owner
            hrefs = el.findall("{DAV:}href")
            self.assertEqual(len(hrefs), 1)
            self.assertEqual(hrefs[0].text, "/principals/user1/")

        asyncio.run(run_test())

    def test_get_value_without_owner(self):
        """Test owner property when resource has no owner."""

        async def run_test():
            prop = access.OwnerProperty()

            class MockResource:
                def get_owner(self):
                    return None

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/resource", resource, el, {})

            # Should have no href elements when no owner
            hrefs = el.findall("{DAV:}href")
            self.assertEqual(len(hrefs), 0)

        asyncio.run(run_test())

    def test_get_value_with_base_href(self):
        """Test owner property with different base href."""

        async def run_test():
            prop = access.OwnerProperty()

            class MockResource:
                def get_owner(self):
                    return "/principals/admin/"

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/calendars/user1/", resource, el, {})

            # Should have href element
            hrefs = el.findall("{DAV:}href")
            self.assertEqual(len(hrefs), 1)
            # Should be absolute or relative to base_href
            self.assertIn("/principals/admin/", hrefs[0].text)

        asyncio.run(run_test())


class GroupMembershipPropertyTests(unittest.TestCase):
    """Tests for GroupMembershipProperty (RFC 3744 Section 4.4)."""

    def test_property_name(self):
        """Test group-membership property name."""
        prop = access.GroupMembershipProperty()
        self.assertEqual(prop.name, "{DAV:}group-membership")

    def test_property_attributes(self):
        """Test group-membership property attributes.

        RFC 3744 Section 4.4: Identifies the groups in which the
        principal is a member.
        """
        prop = access.GroupMembershipProperty()
        self.assertFalse(prop.in_allprops)
        self.assertTrue(prop.live)
        self.assertEqual(prop.resource_type, webdav.PRINCIPAL_RESOURCE_TYPE)

    def test_get_value_with_groups(self):
        """Test group-membership with multiple groups.

        RFC 3744 Section 4.4: Lists all groups the principal is member of.
        """

        async def run_test():
            prop = access.GroupMembershipProperty()

            class MockResource:
                def get_group_membership(self):
                    return [
                        "/principals/groups/staff/",
                        "/principals/groups/developers/",
                    ]

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/principals/user1/", resource, el, {})

            # Should have two href elements
            hrefs = el.findall("{DAV:}href")
            self.assertEqual(len(hrefs), 2)

            # Check both group hrefs are present
            href_texts = [h.text for h in hrefs]
            self.assertIn("/principals/groups/staff/", href_texts)
            self.assertIn("/principals/groups/developers/", href_texts)

        asyncio.run(run_test())

    def test_get_value_no_groups(self):
        """Test group-membership when principal belongs to no groups."""

        async def run_test():
            prop = access.GroupMembershipProperty()

            class MockResource:
                def get_group_membership(self):
                    return []

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/principals/user1/", resource, el, {})

            # Should have no href elements
            hrefs = el.findall("{DAV:}href")
            self.assertEqual(len(hrefs), 0)

        asyncio.run(run_test())

    def test_get_value_single_group(self):
        """Test group-membership with single group."""

        async def run_test():
            prop = access.GroupMembershipProperty()

            class MockResource:
                def get_group_membership(self):
                    return ["/principals/groups/users/"]

            resource = MockResource()
            el = ET.Element("test")

            await prop.get_value("/principals/user1/", resource, el, {})

            # Should have one href element
            hrefs = el.findall("{DAV:}href")
            self.assertEqual(len(hrefs), 1)
            self.assertEqual(hrefs[0].text, "/principals/groups/users/")

        asyncio.run(run_test())


if __name__ == "__main__":
    unittest.main()



# Xandikos
# Copyright (C) 2016-2017 Jelmer Vernooĳ <jelmer@jelmer.uk>, et al.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License or (at your option) any later version of
# the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA  02110-1301, USA.


"""XMPP support.

https://github.com/evert/calendarserver-extensions/blob/master/caldav-pubsubdiscovery.txt
"""

from . import webdav
from .caldav import CALENDAR_RESOURCE_TYPE

ET = webdav.ET


class XmppUriProperty(webdav.Property):
    """xmpp-uri property."""

    name = "{http://calendarserver.org/ns/}xmpp-uri"
    resource_type = CALENDAR_RESOURCE_TYPE
    in_allprops = True
    live = False

    async def get_value(self, base_href, resource, el, environ):
        el.text = resource.get_xmpp_uri()

    async def set_value(self, href, resource, el):
        raise NotImplementedError(self.set_value)


class XmppHeartbeatProperty(webdav.Property):
    """xmpp-heartbeat property."""

    name = "{http://calendarserver.org/ns/}xmpp-heartbeat"
    resource_type = CALENDAR_RESOURCE_TYPE
    in_allprops = True
    live = False

    async def get_value(self, base_href, resource, el, environ):
        (uri, minutes) = resource.get_xmpp_heartbeat()
        uri_el = ET.SubElement(el, "{http://calendarserver.org/ns/}xmpp-heartbeat-uri")
        uri_el.text = uri
        minutes_el = ET.SubElement(
            el, "{http://calendarserver.org/ns/}xmpp-heartbeat-minutes"
        )
        minutes_el.text = str(minutes)

    async def set_value(self, href, resource, el):
        raise NotImplementedError(self.set_value)


class XmppServerProperty(webdav.Property):
    """xmpp-server property."""

    name = "{http://calendarserver.org/ns/}xmpp-server"
    resource_type = CALENDAR_RESOURCE_TYPE
    in_allprops = True
    live = False

    async def get_value(self, base_href, resource, el, environ):
        server = resource.get_xmpp_server()
        el.text = server

    async def set_value(self, href, resource, el):
        raise NotImplementedError(self.set_value)
